from ..logger import Logger
from mona_sdk import MonaSingleMessage
from .mona_client import get_mona_clients
import logging


class MonaLogger(Logger):
    def __init__(
        self,
        mona_creds,
        context_class,
        mona_clients_getter=get_mona_clients,
    ):
        self.client, self.async_client = mona_clients_getter(mona_creds)
        self.context_class = context_class

    def start_monitoring(self, openai_class_name):
        """
        Calls Mona's server to init the given context class specifically for
        the given OpenAI class name.
        """
        response = self.client.create_openai_context_class(
            self.context_class, openai_class_name
        )
        error_message = response.get("error_message")
        if error_message:
            logging.warning(
                f"Problem initializing Mona context class"
                f" '{self.context_class}': {error_message}"
            )
        else:
            logging.info(
                f"Made sure Mona context class '{self.context_class}' "
                "is initialised"
            )
        return response

    def log(self, message, context_id, export_timestamp):
        """
        Logs the given message to Mona.
        """
        return self.client.export(
            MonaSingleMessage(
                message=message,
                contextClass=self.context_class,
                contextId=context_id,
                exportTimestamp=export_timestamp,
            )
        )

    async def alog(self, message, context_id, export_timestamp):
        """
        Async logs the given message to Mona.
        """
        return await self.async_client.export_async(
            MonaSingleMessage(
                message=message,
                contextClass=self.context_class,
                contextId=context_id,
                exportTimestamp=export_timestamp,
            )
        )



import time
from .loggers.mona_logger.mona_logger import MonaLogger
from copy import deepcopy
from types import MappingProxyType

from .exceptions import InvalidLagnchainLLMException
from .endpoints.wrapping_getter import get_endpoint_wrapping
from .loggers.mona_logger.mona_client import get_mona_clients
from .util.func_util import add_conditional_sampling
from .util.async_util import (
    run_in_an_event_loop,
    call_non_blocking_sync_or_async,
)
from .util.openai_util import get_model_param
from .util.tokens_util import get_usage
from .util.stream_util import ResponseGatheringIterator
from .util.validation_util import validate_and_get_sampling_ratio

EMPTY_DICT = MappingProxyType({})

MONA_ARGS_PREFIX = "MONA_"
CONTEXT_ID_ARG_NAME = MONA_ARGS_PREFIX + "context_id"
EXPORT_TIMESTAMP_ARG_NAME = MONA_ARGS_PREFIX + "export_timestamp"
ADDITIONAL_DATA_ARG_NAME = MONA_ARGS_PREFIX + "additional_data"


def _get_logging_message(
    api_name,
    request_input,
    start_time,
    is_exception,
    is_async,
    stream_start_time,
    response,
    analysis_getter,
    message_cleaner,
    additional_data,
):
    """
    Returns a dict object containing all the monitoring analysis to be used
    for data logging.
    """

    message = {
        "input": request_input,
        "latency": time.time() - start_time,
        "stream_start_latency": stream_start_time - start_time
        if stream_start_time is not None
        else None,
        "is_exception": is_exception,
        "api_name": api_name,
        "is_async": is_async,
    }

    if additional_data:
        message["additional_data"] = additional_data

    if response:
        message["response"] = response
        message["analysis"] = analysis_getter(request_input, response)

    return message_cleaner(message)


# TODO(itai): Consider creating some sturct (as NamedTuple or dataclass) for
#   the specs param.


def monitor(
    openai_class,
    mona_creds,
    context_class,
    specs=EMPTY_DICT,
    mona_clients_getter=get_mona_clients,
):
    """
    A simple wrapper around "monitor_with_logger" to use with a Mona logger.
    See "monitor_with_logger" for full documentation.

    Args:
        openai_class: An OpenAI API class to wrap with monitoring
            capabilties.
        mona_creds: Either a dict or pair of Mona API key and secret to
            set up Mona's clients from its SDK
        context_class: The Mona context class name to use for
            monitoring. Use a name of your choice.
        specs: A dictionary of specifications such as monitoring
            sampling ratio.
        mona_clients_getter: Used only for testing purposes.
    """
    return monitor_with_logger(
        openai_class,
        MonaLogger(mona_creds, context_class, mona_clients_getter),
        specs,
    )


def monitor_with_logger(openai_class, logger, specs=EMPTY_DICT):
    """
    Returns a Wrapped version of a given OpenAI class with monitoring logic.

    You can use the returned class' "create" and "acreate" functions
    exactly as you would the original class, and monitoring will be
    taken care of for you.

    This client will automatically monitor for you things like latency,
    prompt and response lengths, number of tokens, etc., along with any
    endpoint parameter usage (e.g., it tracks the "temperature" and
    "max_tokens" params you use).

    The logic for what to do with the calculated analysis data is set by the
    given logger object.

    You can also add other named args when calling "create" or
    "acreate" by using a new named argument called
    "MONA_additional_data" and set it to any JSON serializable
    dictionary.
    This allows you to add metadata about the call such as a prompt
    template ID, information about the context in which the API call is
    made, etc...

    Furthermore, you can add to create/acreate functions mona specific
    arguments:
        MONA_context_id: The unique id of the context in which the call
            is made. By using this ID you can export more data to Mona
            to the same context from other places. If not used, the
            "id" field of the OpenAI Endpoint's response will be used.
        MONA_export_timestamp: Can be used to simulate as if the
            current call was made in a different time, as far as Mona
            is concerned.

    Args:
        openai_class: An OpenAI API class to wrap with monitoring
            capabilties.
        logger: A logger object used to log out the calculated analysis.
        specs: A dictionary of specifications such as monitoring
            sampling ratio.
    """

    sampling_ratio = validate_and_get_sampling_ratio(specs)

    base_class = get_endpoint_wrapping(
        openai_class.__name__, specs
    ).wrap_class(openai_class)

    logger.start_monitoring(openai_class.__name__)

    # TODO(itai): Add call to Mona servers to init the context class if it
    #   isn't inited yet once we have the relevant endpoint for this.

    class MonitoredOpenAI(base_class):
        """
        A monitored version of an openai API class.
        """

        @classmethod
        def _get_logging_message(
            cls,
            kwargs_param,
            start_time,
            is_exception,
            is_async,
            stream_start_time,
            response,
        ):
            """
            Returns a dict to be used for data logging.
            """
            # Recreate the input dict to avoid manipulating the caller's data,
            # and remove Mona-related data.
            request_input = deepcopy(
                {
                    x: kwargs_param[x]
                    for x in kwargs_param
                    if not x.startswith(MONA_ARGS_PREFIX)
                }
            )

            return _get_logging_message(
                api_name=openai_class.__name__,
                request_input=request_input,
                start_time=start_time,
                is_exception=is_exception,
                is_async=is_async,
                stream_start_time=stream_start_time,
                response=response,
                analysis_getter=super()._get_full_analysis,
                message_cleaner=super()._get_clean_message,
                additional_data=kwargs_param.get(ADDITIONAL_DATA_ARG_NAME),
            )

        @classmethod
        async def _inner_create(
            cls, export_function, super_function, args, kwargs
        ):
            """
            The main logic for wrapping create functions with monitoring data
            logging.
            This internal function porovides a template for both sync
            and async activations (helps with wrapping both "create"
            and "acreate").
            """

            is_stream = kwargs.get("stream", False)
            is_async = super_function.__name__ == "acreate"

            response = None

            # will be used only when stream is enabled
            stream_start_time = None

            async def _inner_log_message(is_exception):
                return await call_non_blocking_sync_or_async(
                    export_function,
                    (
                        cls._get_logging_message(
                            kwargs,
                            start_time,
                            is_exception,
                            is_async,
                            stream_start_time,
                            response,
                        ),
                        kwargs.get(
                            CONTEXT_ID_ARG_NAME,
                            response["id"] if response else None,
                        ),
                        kwargs.get(EXPORT_TIMESTAMP_ARG_NAME, start_time),
                    ),
                )

            log_message = add_conditional_sampling(
                _inner_log_message, sampling_ratio
            )

            start_time = time.time()

            async def inner_super_function():
                # Call the actual openai create function without the Mona
                # specific arguments.
                return await call_non_blocking_sync_or_async(
                    super_function,
                    args,
                    {
                        x: kwargs[x]
                        for x in kwargs
                        if not x.startswith(MONA_ARGS_PREFIX)
                    },
                )

            async def inner_handle_exception():
                if not specs.get("avoid_monitoring_exceptions", False):
                    await log_message(True)

            if not is_stream:
                try:
                    response = await inner_super_function()
                except Exception:
                    await inner_handle_exception()
                    raise

                await log_message(False)

                return response

            # From here it's stream handling.

            async def _stream_done_callback(
                final_response, actual_stream_start_time
            ):
                nonlocal response
                nonlocal stream_start_time
                # There is no usage data in returned stream responses, so
                # we add it here.
                response = final_response | {
                    "usage": get_usage(
                        model=get_model_param(kwargs),
                        prompt_texts=base_class._get_all_prompt_texts(kwargs),
                        response_texts=base_class._get_all_response_texts(
                            final_response
                        ),
                    )
                }
                stream_start_time = actual_stream_start_time
                await log_message(False)

            try:
                # Call the actual openai create function without the Mona
                # specific arguments.
                return ResponseGatheringIterator(
                    original_iterator=await inner_super_function(),
                    delta_choice_text_getter=(
                        base_class._get_stream_delta_text_from_choice
                    ),
                    final_choice_getter=base_class._get_final_choice,
                    callback=_stream_done_callback,
                )

            except Exception:
                await inner_handle_exception()
                raise

        @classmethod
        def create(cls, *args, **kwargs):
            """
            A monitored version of the openai base class' "create"
            function.
            """
            return run_in_an_event_loop(
                cls._inner_create(logger.log, super().create, args, kwargs)
            )

        @classmethod
        async def acreate(cls, *args, **kwargs):
            """
            An async monitored version of the openai base class'
            "acreate" function.
            """
            return await cls._inner_create(
                logger.alog, super().acreate, args, kwargs
            )

    return type(base_class.__name__, (MonitoredOpenAI,), {})


def get_rest_monitor(
    openai_endpoint_name,
    mona_creds,
    context_class,
    specs=EMPTY_DICT,
    mona_clients_getter=get_mona_clients,
):
    """
    A wrapper around get_rest_monitor_with_logger that automatically uses
    a Mona logger.
    """
    return get_rest_monitor_with_logger(
        openai_endpoint_name,
        MonaLogger(mona_creds, context_class, mona_clients_getter),
        specs,
    )


def get_rest_monitor_with_logger(
    # TODO(itai): Consider understanding endpoint name from complete url.
    openai_endpoint_name,
    logger,
    specs=EMPTY_DICT,
):
    """
    Returns a client class for monitoring OpenAI REST calls not done
    using the OpenAI python client (e.g., for Azure users using their
    endpoints directly). This isn't a wrapper for any http requesting
    library and doesn't call the OpenAI API for you - it's just an easy
    logging client to log requests, responses and exceptions.
    """

    logger.start_monitoring(openai_endpoint_name)

    sampling_ratio = validate_and_get_sampling_ratio(specs)

    wrapping_logic = get_endpoint_wrapping(openai_endpoint_name, specs)

    class RestClient:
        """
        This will be the returned monitoring class. We follow
        OpenAI's way of doing things by using a static classe with
        relevant class methods.
        """

        @classmethod
        def _inner_log_request(
            cls,
            message_logging_function,
            request_dict,
            additional_data=EMPTY_DICT,
            context_id=None,
            export_timestamp=None,
        ):
            """
            Actual logic for logging requests, responses and exceptions.
            """
            start_time = time.time()

            if additional_data is None:
                additional_data = EMPTY_DICT

            def _inner_log_message(
                is_exception, more_additional_data, response=None
            ):
                return message_logging_function(
                    _get_logging_message(
                        api_name=openai_endpoint_name,
                        request_input=request_dict,
                        start_time=start_time,
                        is_exception=is_exception,
                        is_async=False,
                        # TODO(itai): Support stream in REST as well.
                        stream_start_time=None,
                        response=response,
                        analysis_getter=wrapping_logic.get_full_analysis,
                        message_cleaner=wrapping_logic.get_clean_message,
                        additional_data={
                            **additional_data,
                            **more_additional_data,
                        },
                    ),
                    context_id,
                    export_timestamp,
                )

            log_message = add_conditional_sampling(
                _inner_log_message, sampling_ratio
            )

            def log_response(response, additional_data=EMPTY_DICT):
                """
                Only when this function is called, will data be logged
                out. This function should be called with a
                response object from the OpenAI API as close as
                possible to when it is received to allow accurate
                latency logging.
                """
                return log_message(
                    False,
                    more_additional_data=additional_data,
                    response=response,
                )

            def log_exception(additional_data=EMPTY_DICT):
                return log_message(True, more_additional_data=additional_data)

            return log_response, log_exception

        @classmethod
        def log_request(
            cls,
            request_dict,
            additional_data=None,
            context_id=None,
            export_timestamp=None,
        ):
            """
            Sets up logging for OpenAI request/response objects.

            This function should be called with a request data dict,
            for example, what you would use as "json" when using
            "requests" to post.

            It returns a response logging function to be used with the
            response object, as well as an exception logging function in case
            of exceptions.

            Note that this call does not log anything until one of the
            returned callbacks is called.
            """
            return cls._inner_log_request(
                logger.log,
                request_dict,
                additional_data,
                context_id,
                export_timestamp,
            )

        @classmethod
        def async_log_request(
            cls,
            request_dict,
            additional_data=None,
            context_id=None,
            export_timestamp=None,
        ):
            """
            Async version of "log_request". See function's docstring for more
            details.
            """
            return cls._inner_log_request(
                logger.alog,
                request_dict,
                additional_data,
                context_id,
                export_timestamp,
            )

    return RestClient


def _validate_langchain_llm(llm):
    if not hasattr(llm, "client"):
        raise InvalidLagnchainLLMException(
            "LLM has no client attribute - must be an OpenAI LLM"
        )


def monitor_langchain_llm(
    llm,
    mona_creds,
    context_class,
    specs=EMPTY_DICT,
    mona_clients_getter=get_mona_clients,
):
    """
    Wraps given llm with automatic mona-monitoring logic.
    """
    _validate_langchain_llm(llm)
    llm.client = monitor(
        llm.client, mona_creds, context_class, specs, mona_clients_getter
    )
    return llm


def monitor_langchain_llm_with_logger(llm, logger, specs=EMPTY_DICT):
    """
    Wraps given llm with monitoring logic, logging the analysis with the given
    logger.
    """
    _validate_langchain_llm(llm)
    llm.client = monitor(llm.client, logger, specs)
    return llm



"""
A module to derive text-related metrics such as text length, usage of
specific grammatical words, text repetition, etc...

These analyses can be used to detect significant drifts that could be
caused by hallucinations or bugs.

NOTE: There are many more analyses that can be added here.
"""

from typing import Iterable

PREPOSITIONS = set(
    (
        "aboard",
        "about",
        "above",
        "across",
        "after",
        "against",
        "along",
        "amid",
        "among",
        "around",
        "as",
        "at",
        "before",
        "behind",
        "below",
        "beneath",
        "beside",
        "between",
        "beyond",
        "but",
        "by",
        "concerning",
        "considering",
        "despite",
        "down",
        "during",
        "except",
        "for",
        "from",
        "in",
        "inside",
        "into",
        "like",
        "near",
        "of",
        "off",
        "on",
        "onto",
        "out",
        "outside",
        "over",
        "past",
        "regarding",
        "round",
        "since",
        "through",
        "throughout",
        "till",
        "to",
        "toward",
        "under",
        "underneath",
        "until",
        "unto",
        "up",
        "upon",
        "with",
        "within",
        "without",
    )
)


class TextualAnalyzer:
    """
    An analyzer class that takes a text and provides methods to get analysis
    on that text such as length, word count, etc...
    """

    def __init__(self, text):
        self._text = text
        self._splitted_text = text.split()
        self._prepositions = tuple(
            x for x in self._splitted_text if x in PREPOSITIONS
        )

    def get_length(self):
        """
        Returns the length of the text.
        """
        return len(self._text)

    def get_word_count(self):
        """
        Returns the number of the words in the text.
        """
        return len(self._splitted_text)

    def get_preposition_count(self):
        """
        Returns the number of prepositions in the text.
        """
        return len(self._prepositions)

    def get_preposition_ratio(self):
        """
        Returns the ratio of prepositions in the text.
        """
        word_count = self.get_word_count()
        return self.get_preposition_count() / word_count if word_count else 0

    def get_words_not_in_others_count(
        self, others: Iterable["TextualAnalyzer"]
    ):
        """
        Returns the number of the words in the text that do not appear in the
        given other texts.
        """
        others_words_set = set().union(
            *tuple(other._splitted_text for other in others)
        )
        return len(
            [
                word
                for word in self._splitted_text
                if word not in others_words_set
            ]
        )


def get_textual_analyzers(texts):
    """
    Returns a tuple of TextualAnalyzers for all the given texts.
    """
    return tuple(TextualAnalyzer(text) for text in texts)



from .mona_openai import (
    monitor,
    monitor_with_logger,
    get_rest_monitor,
    get_rest_monitor_with_logger,
    monitor_langchain_llm,
    monitor_langchain_llm_with_logger,
)
from .exceptions import *
from .loggers import *



from .completion import CompletionWrapping, COMPLETION_CLASS_NAME
from .chat_completion import ChatCompletionWrapping, CHAT_COMPLETION_CLASS_NAME
from ..exceptions import WrongOpenAIClassException

# TODO(Itai): This is essetially a nice-looking "switch" statement. We should
#   try to use the name to find the exact monitoring-enrichment function and
#   filename instead of listing all options here.
ENDPOINT_NAME_TO_WRAPPING = {
    COMPLETION_CLASS_NAME: CompletionWrapping,
    CHAT_COMPLETION_CLASS_NAME: ChatCompletionWrapping,
}


def get_endpoint_wrapping(endpoint_name, specs):
    try:
        return ENDPOINT_NAME_TO_WRAPPING[endpoint_name](specs)
    except KeyError:
        raise WrongOpenAIClassException(
            f"Not a supported class name: {endpoint_name}"
        )



from .logger import Logger
from logging import getLogger, INFO


class StandardLogger(Logger):
    """
    A simple logging class that logs monitored data using python's logging
    package.
    """

    def __init__(
        self,
        logging_level=INFO,
        underlying_logger=None,
    ):
        self.underlying_logger = underlying_logger or getLogger("Mona")
        self.level = logging_level

    def log(self, message: dict, context_id=None, export_timestamp=None):
        self.underlying_logger.log(
            self.level,
            {
                "message": message,
                "context_id": context_id,
                "export_timestamp": export_timestamp,
            },
        )

    async def alog(
        self, message: dict, context_id=None, export_timestamp=None
    ):
        return self.log(message, context_id, export_timestamp)

    def start_monitoring(self, openai_class_name):
        self.underlying_logger.log(
            self.level, f"Started monitoring for OpenAI's {openai_class_name}"
        )



"""
Tests for Completion api Mona wrapping.

NOTE: Many of these tests can be considered more generic than just for
    "Completion", since they test general mona-openai capabilities
    (e.g., sampling ratio, basic flows). We should either keep them
    here (since they do use the "Completion" API after all) or create a
    more generic test module in addition to this one
"""
import asyncio
from copy import deepcopy

import pytest
from openai import Completion

from mona_openai.exceptions import InvalidSamplingRatioException
from mona_openai.mona_openai import (
    CONTEXT_ID_ARG_NAME,
    EXPORT_TIMESTAMP_ARG_NAME,
    monitor,
    get_rest_monitor,
)
from .mocks.mock_openai import (
    mockCreateException,
    mockCreateExceptionCommand,
    get_mock_openai_class,
)
from .mocks.mock_mona_client import get_mock_mona_clients_getter

_DEFAULT_CONTEXT_CLASS = "TEST_CLASS"

_DEFAULT_RESPONSE_TEXT = "\n\nMy name is"

_DEFAULT_RESPONSE_COMMON_VARIABLES = {
    "created": 1679231055,
    "id": "cmpl-6vmzn6DUc2ZNjkyEvAyTf2tAgPl3A",
    "model": "text-ada-001",
    "object": "text_completion",
}

_DEFAULT_RESPONSE = {
    "choices": [
        {
            "finish_reason": "length",
            "index": 0,
            "text": _DEFAULT_RESPONSE_TEXT,
        }
    ],
    "usage": {"completion_tokens": 5, "prompt_tokens": 8, "total_tokens": 13},
} | _DEFAULT_RESPONSE_COMMON_VARIABLES


def _get_response_without_texts(response):
    new_response = deepcopy(response)
    for choice in new_response["choices"]:
        choice.pop("text")
    return new_response


_DEFAULT_EXPORTED_RESPONSE = _get_response_without_texts(_DEFAULT_RESPONSE)

_DEFAULT_INPUT = {
    "prompt": "I want to generate some text about ",
    "model": "text-ada-001",
    "temperature": 0.6,
    "n": 1,
    "max_tokens": 5,
}

# By default we don't export the prompt to Mona
_DEFAULT_EXPORTED_INPUT = {
    x: _DEFAULT_INPUT[x] for x in _DEFAULT_INPUT if x != "prompt"
}

_DEFAULT_ANALYSIS = {
    "privacy": {
        "prompt_phone_number_count": (0,),
        "answer_unknown_phone_number_count": (0,),
        "prompt_email_count": (0,),
        "answer_unknown_email_count": (0,),
    },
    "textual": {
        "prompt_length": (35,),
        "answer_length": (12,),
        "prompt_word_count": (7,),
        "answer_word_count": (3,),
        "prompt_preposition_count": (2,),
        "prompt_preposition_ratio": (0.2857142857142857,),
        "answer_preposition_count": (0,),
        "answer_preposition_ratio": (0.0,),
        "answer_words_not_in_prompt_count": (3,),
        "answer_words_not_in_prompt_ratio": (1.0,),
    },
    "profanity": {
        "prompt_profanity_prob": (0.05,),
        "answer_profanity_prob": (0.05,),
        "prompt_has_profanity": (False,),
        "answer_has_profanity": (False,),
    },
}


def _remove_none_values(dict):
    return {x: y for x, y in dict.items() if y is not None}


def _get_mock_openai_class(*args, **kwargs):
    return get_mock_openai_class(Completion, *args, **kwargs)


def _get_mona_message(
    input=_DEFAULT_EXPORTED_INPUT,
    is_exception=False,
    is_async=False,
    is_stream=None,
    response=_DEFAULT_EXPORTED_RESPONSE,
    analysis=_DEFAULT_ANALYSIS,
    context_class=_DEFAULT_CONTEXT_CLASS,
    context_id=None,
    export_timestamp=None,
    additional_data=None,
):
    message = {
        "message": {
            "input": input,
            "is_exception": is_exception,
            "api_name": "Completion",
            "is_async": is_async,
            "response": response,
            "analysis": analysis,
            "additional_data": additional_data,
        },
        "context_class": context_class,
        "context_id": context_id,
        "export_timestamp": export_timestamp,
    }

    message["message"] = _remove_none_values(message["message"])
    return _remove_none_values(message)


def test_basic():
    monitor(
        _get_mock_openai_class((_DEFAULT_RESPONSE,), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (_get_mona_message(),), ()
        ),
    ).create(**_DEFAULT_INPUT)


def test_multiple_prompts():
    new_input = deepcopy(_DEFAULT_INPUT)
    new_input["prompt"] = (
        "I want to generate some text about ",
        "I also want to generate some text about ",
    )
    expected_input = deepcopy(new_input)
    expected_input.pop("prompt")

    new_response = deepcopy(_DEFAULT_RESPONSE)
    new_response["choices"] = [
        {
            "finish_reason": "length",
            "index": 0,
            "logprobs": None,
            "text": "\n\nMy name is",
        },
        {
            "finish_reason": "length",
            "index": 1,
            "logprobs": None,
            "text": "\n\nMy thing is",
        },
    ]
    new_expected_response = _get_response_without_texts(new_response)

    new_analysis = {
        "privacy": {
            "prompt_phone_number_count": (0, 0),
            "answer_unknown_phone_number_count": (0, 0),
            "prompt_email_count": (0, 0),
            "answer_unknown_email_count": (0, 0),
        },
        "textual": {
            "prompt_length": (35, 40),
            "answer_length": (12, 13),
            "prompt_word_count": (7, 8),
            "answer_word_count": (3, 3),
            "prompt_preposition_count": (2, 2),
            "prompt_preposition_ratio": (0.2857142857142857, 0.25),
            "answer_preposition_count": (0, 0),
            "answer_preposition_ratio": (0.0, 0.0),
            "answer_words_not_in_prompt_count": (3, 3),
            "answer_words_not_in_prompt_ratio": (1.0, 1.0),
        },
        "profanity": {
            "prompt_profanity_prob": (0.05, 0.05),
            "answer_profanity_prob": (0.05, 0.01),
            "prompt_has_profanity": (False, False),
            "answer_has_profanity": (False, False),
        },
    }

    monitor(
        _get_mock_openai_class((new_response,), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (
                _get_mona_message(
                    response=new_expected_response,
                    input=expected_input,
                    analysis=new_analysis,
                ),
            ),
            (),
        ),
    ).create(**new_input)


def test_rest():
    get_rest_monitor(
        Completion.__name__,
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (_get_mona_message(),), ()
        ),
    ).log_request(_DEFAULT_INPUT)[0](_DEFAULT_RESPONSE)


def test_rest_more_additional_data():
    additional_data = {"foo": "bar"}
    more_additional_data = {"foo2": "bar2"}
    total_additional_data = {**additional_data, **more_additional_data}
    get_rest_monitor(
        Completion.__name__,
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (_get_mona_message(additional_data=total_additional_data),), ()
        ),
    ).log_request(_DEFAULT_INPUT, additional_data=additional_data)[0](
        _DEFAULT_RESPONSE, additional_data=more_additional_data
    )


def test_rest_async():
    asyncio.run(
        get_rest_monitor(
            Completion.__name__,
            (),
            _DEFAULT_CONTEXT_CLASS,
            mona_clients_getter=get_mock_mona_clients_getter(
                (), (_get_mona_message(),)
            ),
        ).async_log_request(_DEFAULT_INPUT)[0](_DEFAULT_RESPONSE)
    )


def test_rest_exception():
    get_rest_monitor(
        Completion.__name__,
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (
                _get_mona_message(
                    is_exception=True, response=None, analysis=None
                ),
            ),
            (),
        ),
    ).log_request(_DEFAULT_INPUT)[1]()


def test_rest_exception_async():
    asyncio.run(
        get_rest_monitor(
            Completion.__name__,
            (),
            _DEFAULT_CONTEXT_CLASS,
            mona_clients_getter=get_mock_mona_clients_getter(
                (),
                (
                    _get_mona_message(
                        is_exception=True, response=None, analysis=None
                    ),
                ),
            ),
        ).async_log_request(_DEFAULT_INPUT)[1]()
    )


def test_export_response_text():
    monitor(
        _get_mock_openai_class((_DEFAULT_RESPONSE,), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        {"export_response_texts": True},
        mona_clients_getter=get_mock_mona_clients_getter(
            (_get_mona_message(response=_DEFAULT_RESPONSE),), ()
        ),
    ).create(**_DEFAULT_INPUT)


def test_export_prompt():
    monitor(
        _get_mock_openai_class((_DEFAULT_RESPONSE,), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        {"export_prompt": True},
        mona_clients_getter=get_mock_mona_clients_getter(
            (_get_mona_message(input=_DEFAULT_INPUT),), ()
        ),
    ).create(**_DEFAULT_INPUT)


def test_bad_sampling_ratios():
    with pytest.raises(InvalidSamplingRatioException):
        monitor(
            _get_mock_openai_class((_DEFAULT_RESPONSE,), ()),
            (),
            _DEFAULT_CONTEXT_CLASS,
            {"sampling_ratio": 1.1},
            mona_clients_getter=get_mock_mona_clients_getter(
                (_get_mona_message(),), ()
            ),
        )

    with pytest.raises(InvalidSamplingRatioException):
        monitor(
            _get_mock_openai_class((_DEFAULT_RESPONSE,), ()),
            (),
            _DEFAULT_CONTEXT_CLASS,
            {"sampling_ratio": -1},
            mona_clients_getter=get_mock_mona_clients_getter(
                (_get_mona_message(),), ()
            ),
        )


def test_async():
    monitored_completion = monitor(
        _get_mock_openai_class((), (_DEFAULT_RESPONSE,)),
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (), (_get_mona_message(is_async=True),)
        ),
    )

    asyncio.run(monitored_completion.acreate(**_DEFAULT_INPUT))


def test_exception():
    monitored_completion = monitor(
        _get_mock_openai_class((mockCreateExceptionCommand(),), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (
                _get_mona_message(
                    is_exception=True, response=None, analysis=None
                ),
            ),
            (),
        ),
    )

    with pytest.raises(mockCreateException):
        monitored_completion.create(**_DEFAULT_INPUT)


def test_exception_without_monitoring():
    monitored_completion = monitor(
        _get_mock_openai_class((mockCreateExceptionCommand(),), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        {"avoid_monitoring_exceptions": True},
        mona_clients_getter=get_mock_mona_clients_getter((), ()),
    )

    with pytest.raises(mockCreateException):
        monitored_completion.create(**_DEFAULT_INPUT)


def test_context_id():
    context_id = "some_context_id"
    monitor(
        _get_mock_openai_class((_DEFAULT_RESPONSE,), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (_get_mona_message(context_id=context_id),), ()
        ),
    ).create(**{**_DEFAULT_INPUT, CONTEXT_ID_ARG_NAME: context_id})


def test_export_timestamp():
    export_timestamp = 1679244447
    monitor(
        _get_mock_openai_class((_DEFAULT_RESPONSE,), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (_get_mona_message(export_timestamp=export_timestamp),), ()
        ),
    ).create(**{**_DEFAULT_INPUT, EXPORT_TIMESTAMP_ARG_NAME: export_timestamp})


def test_no_profanity():
    expected_analysis = deepcopy(_DEFAULT_ANALYSIS)
    expected_analysis.pop("profanity")
    monitor(
        _get_mock_openai_class((_DEFAULT_RESPONSE,), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        {"analysis": {"profanity": False}},
        mona_clients_getter=get_mock_mona_clients_getter(
            (_get_mona_message(analysis=expected_analysis),), ()
        ),
    ).create(**_DEFAULT_INPUT)


def test_no_textual_or_privacy():
    expected_analysis = deepcopy(_DEFAULT_ANALYSIS)
    expected_analysis.pop("privacy")
    expected_analysis.pop("textual")
    monitor(
        _get_mock_openai_class((_DEFAULT_RESPONSE,), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        {"analysis": {"privacy": False, "textual": False}},
        mona_clients_getter=get_mock_mona_clients_getter(
            (_get_mona_message(analysis=expected_analysis),), ()
        ),
    ).create(**_DEFAULT_INPUT)


def test_multiple_answers():
    new_input = deepcopy(_DEFAULT_INPUT)
    new_input["n"] = 3
    expected_input = deepcopy(new_input)
    expected_input.pop("prompt")

    new_response = deepcopy(_DEFAULT_RESPONSE)
    new_response["choices"] = [
        {
            "finish_reason": "length",
            "index": 0,
            "logprobs": None,
            "text": "\n\nMy name is",
        },
        {
            "finish_reason": "length",
            "index": 1,
            "logprobs": None,
            "text": "\n\nMy thing is",
        },
        {
            "finish_reason": "length",
            "index": 2,
            "logprobs": None,
            "text": "\n\nbladf",
        },
    ]

    new_expected_response = _get_response_without_texts(new_response)

    new_analysis = {
        "privacy": {
            "prompt_phone_number_count": (0,),
            "answer_unknown_phone_number_count": (0, 0, 0),
            "prompt_email_count": (0,),
            "answer_unknown_email_count": (0, 0, 0),
        },
        "textual": {
            "prompt_length": (35,),
            "answer_length": (12, 13, 7),
            "prompt_word_count": (7,),
            "answer_word_count": (3, 3, 1),
            "prompt_preposition_count": (2,),
            "prompt_preposition_ratio": (0.2857142857142857,),
            "answer_preposition_count": (0, 0, 0),
            "answer_preposition_ratio": (0.0, 0.0, 0.0),
            "answer_words_not_in_prompt_count": (3, 3, 1),
            "answer_words_not_in_prompt_ratio": (1.0, 1.0, 1.0),
        },
        "profanity": {
            "prompt_profanity_prob": (0.05,),
            "answer_profanity_prob": (0.05, 0.01, 0.05),
            "prompt_has_profanity": (False,),
            "answer_has_profanity": (False, False, False),
        },
    }

    monitor(
        _get_mock_openai_class((new_response,), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (
                _get_mona_message(
                    response=new_expected_response,
                    input=expected_input,
                    analysis=new_analysis,
                ),
            ),
            (),
        ),
    ).create(**new_input)


def test_additional_data():
    new_input = deepcopy(_DEFAULT_INPUT)
    additional_data = {"foo": "bar", "foo2": 2}
    new_input["MONA_additional_data"] = additional_data

    monitor(
        _get_mock_openai_class((_DEFAULT_RESPONSE,), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (_get_mona_message(additional_data=additional_data),), ()
        ),
    ).create(**new_input)


def test_stream():
    def response_generator():
        words = _DEFAULT_RESPONSE_TEXT.split(" ")
        last_index = len(words) - 1
        for i, word in enumerate(words):
            choice = {
                "text": (word + " ") if i < last_index else word,
                "index": 0,
                "logprobs": None,
                "finish_reason": None if i < last_index else "length",
            }
            yield _DEFAULT_RESPONSE_COMMON_VARIABLES | {"choices": [choice]}

    input = deepcopy(_DEFAULT_INPUT)
    input["stream"] = True

    expected_input = deepcopy(input)
    expected_input.pop("prompt")

    for _ in monitor(
        _get_mock_openai_class((response_generator(),), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (_get_mona_message(is_stream=True, input=expected_input),), ()
        ),
    ).create(**input):
        pass


def test_stream_multiple_answers():
    def response_generator():
        words = _DEFAULT_RESPONSE_TEXT.split(" ")
        for i, word in enumerate(words):
            yield _DEFAULT_RESPONSE_COMMON_VARIABLES | {
                "choices": [
                    {
                        "text": (word + " ") if i < len(words) - 1 else word,
                        "index": 0,
                        "logprobs": None,
                        "finish_reason": None
                        if i < len(words) - 1
                        else "length",
                    }
                ]
            }
            yield _DEFAULT_RESPONSE_COMMON_VARIABLES | {
                "choices": [
                    {
                        "text": (word + " ") if i < len(words) - 1 else word,
                        "index": 1,
                        "logprobs": None,
                        "finish_reason": None
                        if i < len(words) - 1
                        else "length",
                    }
                ]
            }

    input = deepcopy(_DEFAULT_INPUT)
    input["stream"] = True
    input["n"] = 2

    expected_input = deepcopy(input)
    expected_input.pop("prompt")

    expected_response = deepcopy(_DEFAULT_EXPORTED_RESPONSE)
    expected_response["choices"] += deepcopy(expected_response["choices"])
    expected_response["choices"][1]["index"] = 1
    expected_response["usage"] = {
        "completion_tokens": 10,
        "prompt_tokens": 8,
        "total_tokens": 18,
    }

    new_analysis = {
        "privacy": {
            "prompt_phone_number_count": (0,),
            "answer_unknown_phone_number_count": (0, 0),
            "prompt_email_count": (0,),
            "answer_unknown_email_count": (0, 0),
        },
        "textual": {
            "prompt_length": (35,),
            "answer_length": (12, 12),
            "prompt_word_count": (7,),
            "answer_word_count": (3, 3),
            "prompt_preposition_count": (2,),
            "prompt_preposition_ratio": (0.2857142857142857,),
            "answer_preposition_count": (0, 0),
            "answer_preposition_ratio": (0.0, 0.0),
            "answer_words_not_in_prompt_count": (3, 3),
            "answer_words_not_in_prompt_ratio": (1.0, 1.0),
        },
        "profanity": {
            "prompt_profanity_prob": (0.05,),
            "answer_profanity_prob": (0.05, 0.05),
            "prompt_has_profanity": (False,),
            "answer_has_profanity": (False, False),
        },
    }

    for _ in monitor(
        _get_mock_openai_class((response_generator(),), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (
                _get_mona_message(
                    is_stream=True,
                    input=expected_input,
                    response=expected_response,
                    analysis=new_analysis,
                ),
            ),
            (),
        ),
    ).create(**input):
        pass


def test_stream_async():
    async def response_generator():
        words = _DEFAULT_RESPONSE_TEXT.split(" ")
        for i, word in enumerate(words):
            choice = {
                "text": (word + " ") if i < len(words) - 1 else word,
                "index": 0,
                "logprobs": None,
                "finish_reason": None if i < len(words) - 1 else "length",
            }
            yield _DEFAULT_RESPONSE_COMMON_VARIABLES | {"choices": [choice]}

    input = deepcopy(_DEFAULT_INPUT)
    input["stream"] = True

    expected_input = deepcopy(input)
    expected_input.pop("prompt")

    async def iterate_gen():
        async for _ in await monitor(
            _get_mock_openai_class((), (response_generator(),)),
            (),
            _DEFAULT_CONTEXT_CLASS,
            mona_clients_getter=get_mock_mona_clients_getter(
                (),
                (
                    _get_mona_message(
                        is_stream=True, is_async=True, input=expected_input
                    ),
                ),
            ),
        ).acreate(**input):
            pass

    asyncio.run(iterate_gen())



from .logger import Logger
from .in_memory_logging import InMemoryLogger
from .standard_logging import StandardLogger
from .file_logger import FileLogger
from .mona_logger.mona_logger import MonaLogger



import abc


class Logger(metaclass=abc.ABCMeta):
    """
    An abstract class/interface for logging messages containing OpenAI calls
    analysis data.
    """

    def start_monitoring(self, openai_class_name):
        """
        This function will be called once this logger is used for wrapping
        an OpenAI class.
        Child classes may choose to use this hook in order to run some logic
        preliminary to actual logging.
        """
        pass

    @abc.abstractmethod
    def log(self, message: dict, context_id=None, export_timestamp=None):
        """
        Every child class must implement this basic function which gets a
        dictionary to be logged.

        The interface here allows the logger to get two other parameters:
        - context_id: used to trace different logs related to the same
            context, in logging mechanisms where such a capability is
            relevant.
        - export_timestamp: Used to simulate logging of historical data,
            allowing the caller to specify when the message was created, in
            logging mechanisms where such a capability is relevant.
        """
        pass

    @abc.abstractmethod
    async def alog(
        self, message: dict, context_id=None, export_timestamp=None
    ):
        """
        Child classes should implement this async version of the "log"
        function.
        """
        pass



from .logger import Logger
from collections import deque

DEFAULT_MAX_LEN = 1000


class InMemoryLogger(Logger):
    """
    A simple logging class that saves monitored data in an in-memory list
    under self.latest_messages.
    """

    def __init__(self, max_len=DEFAULT_MAX_LEN):
        self.latest_messages = deque(maxlen=max_len)

    def log(self, message: dict, context_id=None, export_timestamp=None):
        self.latest_messages.append(
            {
                "message": message,
                "context_id": context_id,
                "export_timestamp": export_timestamp,
            }
        )

    async def alog(
        self, message: dict, context_id=None, export_timestamp=None
    ):
        self.log(message, context_id, export_timestamp)



class WrongOpenAIClassException(Exception):
    pass


class InvalidSamplingRatioException(Exception):
    pass


class InvalidMonaCredsException(Exception):
    pass


class InvalidLagnchainLLMException(Exception):
    pass



"""
The Mona wrapping code for OpenAI's ChatCompletion API.
"""
from copy import deepcopy
from functools import wraps

from ..util.oop_util import create_combined_object
from ..analysis.privacy import get_privacy_analyzers
from ..analysis.profanity import get_profanity_prob, get_has_profanity
from ..analysis.textual import get_textual_analyzers
from .endpoint_wrapping import OpenAIEndpointWrappingLogic

CHAT_COMPLETION_CLASS_NAME = "ChatCompletion"


def _get_choices_texts(response):
    return tuple(
        (choice["message"]["content"] for choice in response["choices"])
    )


def _get_prompt_texts(request):
    return tuple(message["content"] for message in request["messages"])


def _get_texts(func):
    def wrapper(self, input, response):
        return func(
            self,
            input["messages"][-1]["content"]
            if input["messages"][-1]["role"] == "user"
            else None,
            _get_prompt_texts(input),
            _get_choices_texts(response),
        )

    return wrapper


def _get_analyzers(analyzers_getter):
    def decorator(func):
        @wraps(func)
        def wrapper(self, last_user_message, messages, answers):
            return func(
                self,
                analyzers_getter((last_user_message,))[0]
                if last_user_message is not None
                else None,
                analyzers_getter(messages),
                analyzers_getter(answers),
            )

        return wrapper

    return decorator


class ChatCompletionWrapping(OpenAIEndpointWrappingLogic):
    def _get_endpoint_name(self):
        return CHAT_COMPLETION_CLASS_NAME

    def get_clean_message(self, message):
        """
        Returns a copy of the given message with relevant data removed, for
        example the actual texts, to avoid sending such information, that
        is sometimes sensitive, to Mona.
        """
        new_message = deepcopy(message)
        if not self._specs.get("export_prompt", False):
            for input_message in new_message["input"]["messages"]:
                input_message.pop("content", None)

        if "response" in message and not self._specs.get(
            "export_response_texts", False
        ):
            for choice in new_message["response"]["choices"]:
                choice["message"].pop("content", None)

        return new_message

    @_get_texts
    @_get_analyzers(get_privacy_analyzers)
    def _get_full_privacy_analysis(
        self,
        last_user_message_analyzer,
        messages_privacy_analyzers,
        answers_privacy_analyzers,
    ):
        combined_messages = create_combined_object(messages_privacy_analyzers)
        combined_answers = create_combined_object(answers_privacy_analyzers)
        ret = {
            "total_prompt_phone_number_count": sum(
                combined_messages.get_phone_numbers_count()
            ),
            "answer_unknown_phone_number_count": (
                combined_answers.get_previously_unseen_phone_numbers_count(
                    messages_privacy_analyzers
                )
            ),
            "total_prompt_email_count": sum(
                combined_messages.get_emails_count()
            ),
            "answer_unknown_email_count": (
                combined_answers.get_previously_unseen_emails_count(
                    messages_privacy_analyzers
                )
            ),
        }
        if last_user_message_analyzer is not None:
            ret.update(
                {
                    "last_user_message_phone_number_count": (
                        last_user_message_analyzer.get_phone_numbers_count()
                    ),
                    "last_user_message_emails_count": (
                        last_user_message_analyzer.get_emails_count()
                    ),
                }
            )
        return ret

    @_get_texts
    @_get_analyzers(get_textual_analyzers)
    def _get_full_textual_analysis(
        self,
        last_user_message_analyzer,
        messages_textual_analyzers,
        answers_textual_analyzers,
    ):
        combined_messages = create_combined_object(messages_textual_analyzers)
        combined_answers = create_combined_object(answers_textual_analyzers)
        total_prompt_word_count = sum(combined_messages.get_word_count())
        total_prompt_preposition_count = sum(
            combined_messages.get_preposition_count()
        )

        ret = {
            "total_prompt_length": sum(combined_messages.get_length()),
            "answer_length": combined_answers.get_length(),
            "total_prompt_word_count": total_prompt_word_count,
            "answer_word_count": combined_answers.get_word_count(),
            "total_prompt_preposition_count": total_prompt_preposition_count,
            "total_prompt_preposition_ratio": total_prompt_preposition_count
            / total_prompt_word_count
            if total_prompt_word_count != 0
            else None,
            "answer_preposition_count": (
                combined_answers.get_preposition_count()
            ),
            "answer_preposition_ratio": (
                combined_answers.get_preposition_ratio()
            ),
            "answer_words_not_in_prompt_count": (
                combined_answers.get_words_not_in_others_count(
                    messages_textual_analyzers
                )
            ),
            "answer_words_not_in_prompt_ratio": tuple(
                analyzer.get_words_not_in_others_count(
                    messages_textual_analyzers
                )
                / analyzer.get_word_count()
                if analyzer.get_word_count() > 0
                else 0.0
                for analyzer in answers_textual_analyzers
            ),
        }

        if last_user_message_analyzer is not None:
            ret.update(
                {
                    "last_user_message_length": (
                        last_user_message_analyzer.get_length()
                    ),
                    "last_user_message_word_count": (
                        last_user_message_analyzer.get_word_count()
                    ),
                    "last_user_message_preposition_count": (
                        last_user_message_analyzer.get_preposition_count()
                    ),
                    "last_user_message_preposition_ratio": (
                        last_user_message_analyzer.get_preposition_ratio()
                    ),
                }
            )

        return ret

    @_get_texts
    def _get_full_profainty_analysis(
        self, last_user_message, messages, answers
    ):
        ret = {
            "prompt_profanity_prob": get_profanity_prob(messages),
            "prompt_has_profanity": get_has_profanity(messages),
            "answer_profanity_prob": get_profanity_prob(answers),
            "answer_has_profanity": get_has_profanity(answers),
        }

        if last_user_message is not None:
            ret.update(
                {
                    "last_user_message_profanity_prob": get_profanity_prob(
                        (last_user_message,)
                    )[0],
                    "last_user_message_has_profanity": get_has_profanity(
                        (last_user_message,)
                    )[0],
                }
            )

        return ret

    def get_stream_delta_text_from_choice(self, choice):
        return choice["delta"].get("content", "")

    def get_final_choice(self, text):
        return {"message": {"role": "assistant", "content": text}}

    def get_all_prompt_texts(self, request):
        return _get_prompt_texts(request)

    def get_all_response_texts(self, response):
        return _get_choices_texts(response)



"""
Tests for ChatCompletion api Mona wrapping.
"""
from copy import deepcopy

from openai import ChatCompletion

from mona_openai.mona_openai import (
    monitor,
    get_rest_monitor,
)
from .mocks.mock_openai import (
    get_mock_openai_class,
)
from .mocks.mock_mona_client import get_mock_mona_clients_getter

_DEFAULT_CONTEXT_CLASS = "TEST_CLASS"

_DEFAULT_RESPONSE_TEXT = "\n\nMy name is"

_DEFAULT_RESPONSE_COMMON_VARIABLES = {
    "created": 1684827250,
    "id": "chatcmpl-7JGp0PUipNwDQeja4P7SwSLa1I19H",
    "model": "gpt-3.5-turbo-0301",
    "object": "chat.completion",
}

_DEFAULT_RESPONSE = {
    "choices": [
        {
            "finish_reason": "length",
            "index": 0,
            "message": {
                "role": "assistant",
                "content": _DEFAULT_RESPONSE_TEXT,
            },
        }
    ],
    "usage": {
        "completion_tokens": 4,
        "prompt_tokens": 8,
        "total_tokens": 12,
    },
} | _DEFAULT_RESPONSE_COMMON_VARIABLES


def _get_response_without_texts(response):
    new_response = deepcopy(response)
    for choice in new_response["choices"]:
        choice["message"].pop("content")
    return new_response


_DEFAULT_EXPORTED_RESPONSE = _get_response_without_texts(_DEFAULT_RESPONSE)

_DEFAULT_INPUT = {
    "model": "gpt-3.5-turbo",
    "messages": [
        {"role": "user", "content": "I want to generate some text about "}
    ],
    "max_tokens": 20,
    "n": 1,
    "temperature": 0.2,
}


# By default we don't export the prompt to Mona
def _remove_text_content_from_input(input):
    new_input = deepcopy(input)
    for message in new_input["messages"]:
        message.pop("content", None)

    return new_input


_DEFAULT_EXPORTED_INPUT = _remove_text_content_from_input(_DEFAULT_INPUT)

_DEFAULT_ANALYSIS = {
    "privacy": {
        "total_prompt_phone_number_count": 0,
        "answer_unknown_phone_number_count": (0,),
        "total_prompt_email_count": 0,
        "answer_unknown_email_count": (0,),
        "last_user_message_phone_number_count": 0,
        "last_user_message_emails_count": 0,
    },
    "textual": {
        "total_prompt_length": 35,
        "answer_length": (12,),
        "total_prompt_word_count": 7,
        "answer_word_count": (3,),
        "total_prompt_preposition_count": 2,
        "total_prompt_preposition_ratio": 0.2857142857142857,
        "answer_preposition_count": (0,),
        "answer_preposition_ratio": (0,),
        "answer_words_not_in_prompt_count": (3,),
        "answer_words_not_in_prompt_ratio": (1,),
        "last_user_message_length": 35,
        "last_user_message_word_count": 7,
        "last_user_message_preposition_count": 2,
        "last_user_message_preposition_ratio": 0.2857142857142857,
    },
    "profanity": {
        "prompt_profanity_prob": (0.05,),
        "prompt_has_profanity": (False,),
        "answer_profanity_prob": (0.05,),
        "answer_has_profanity": (False,),
        "last_user_message_profanity_prob": 0.05,
        "last_user_message_has_profanity": False,
    },
}


def _remove_none_values(dict):
    return {x: y for x, y in dict.items() if y is not None}


def _get_mock_openai_class(*args, **kwargs):
    return get_mock_openai_class(ChatCompletion, *args, **kwargs)


def _get_mona_message(
    input=_DEFAULT_EXPORTED_INPUT,
    is_exception=False,
    is_async=False,
    is_stream=None,
    response=_DEFAULT_EXPORTED_RESPONSE,
    analysis=_DEFAULT_ANALYSIS,
    context_class=_DEFAULT_CONTEXT_CLASS,
    context_id=None,
    export_timestamp=None,
    additional_data=None,
):
    message = {
        "message": {
            "input": input,
            "is_exception": is_exception,
            "api_name": "ChatCompletion",
            "is_async": is_async,
            "response": response,
            "analysis": analysis,
            "additional_data": additional_data,
        },
        "context_class": context_class,
        "context_id": context_id,
        "export_timestamp": export_timestamp,
    }

    message["message"] = _remove_none_values(message["message"])
    return _remove_none_values(message)


def test_basic():
    monitor(
        _get_mock_openai_class((_DEFAULT_RESPONSE,), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (_get_mona_message(),), ()
        ),
    ).create(**_DEFAULT_INPUT)


def test_multiple_messages_not_ending_with_user_message():
    new_input = deepcopy(_DEFAULT_INPUT)
    new_input["messages"] = (
        [{"role": "system", "content": "you are an assistant"}]
        + new_input["messages"]
        + [{"role": "assistant", "content": "some initial answer"}]
    )

    expected_input = _remove_text_content_from_input(new_input)

    new_analysis = {
        "privacy": {
            "total_prompt_phone_number_count": 0,
            "answer_unknown_phone_number_count": (0,),
            "total_prompt_email_count": 0,
            "answer_unknown_email_count": (0,),
        },
        "textual": {
            "total_prompt_length": 74,
            "answer_length": (12,),
            "total_prompt_word_count": 14,
            "answer_word_count": (3,),
            "total_prompt_preposition_count": 2,
            "total_prompt_preposition_ratio": 0.14285714285714285,
            "answer_preposition_count": (0,),
            "answer_preposition_ratio": (0,),
            "answer_words_not_in_prompt_count": (3,),
            "answer_words_not_in_prompt_ratio": (1,),
        },
        "profanity": {
            "prompt_profanity_prob": (0.05, 0.05, 0.02),
            "prompt_has_profanity": (False, False, False),
            "answer_profanity_prob": (0.05,),
            "answer_has_profanity": (False,),
        },
    }

    monitor(
        _get_mock_openai_class((_DEFAULT_RESPONSE,), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (
                _get_mona_message(
                    input=expected_input,
                    analysis=new_analysis,
                ),
            ),
            (),
        ),
    ).create(**new_input)


def test_multiple_messages():
    new_input = deepcopy(_DEFAULT_INPUT)
    new_input["messages"] = (
        [{"role": "system", "content": "you are an assistant"}]
        + new_input["messages"]
        + [
            {"role": "assistant", "content": "some initial answer"},
            {"role": "user", "content": "some user new prompt"},
        ]
    )

    expected_input = _remove_text_content_from_input(new_input)

    new_analysis = {
        "privacy": {
            "total_prompt_phone_number_count": 0,
            "answer_unknown_phone_number_count": (0,),
            "total_prompt_email_count": 0,
            "answer_unknown_email_count": (0,),
            "last_user_message_phone_number_count": 0,
            "last_user_message_emails_count": 0,
        },
        "textual": {
            "total_prompt_length": 94,
            "answer_length": (12,),
            "total_prompt_word_count": 18,
            "answer_word_count": (3,),
            "total_prompt_preposition_count": 2,
            "total_prompt_preposition_ratio": 0.1111111111111111,
            "answer_preposition_count": (0,),
            "answer_preposition_ratio": (0,),
            "answer_words_not_in_prompt_count": (3,),
            "answer_words_not_in_prompt_ratio": (1,),
            "last_user_message_length": 20,
            "last_user_message_word_count": 4,
            "last_user_message_preposition_count": 0,
            "last_user_message_preposition_ratio": 0,
        },
        "profanity": {
            "prompt_profanity_prob": (0.05, 0.05, 0.02, 0.03),
            "prompt_has_profanity": (False, False, False, False),
            "answer_profanity_prob": (0.05,),
            "answer_has_profanity": (False,),
            "last_user_message_profanity_prob": 0.03,
            "last_user_message_has_profanity": False,
        },
    }

    monitor(
        _get_mock_openai_class((_DEFAULT_RESPONSE,), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (
                _get_mona_message(
                    input=expected_input,
                    analysis=new_analysis,
                ),
            ),
            (),
        ),
    ).create(**new_input)


def test_rest():
    get_rest_monitor(
        ChatCompletion.__name__,
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (_get_mona_message(),), ()
        ),
    ).log_request(_DEFAULT_INPUT)[0](_DEFAULT_RESPONSE)


def test_rest_exception():
    get_rest_monitor(
        ChatCompletion.__name__,
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (
                _get_mona_message(
                    is_exception=True, response=None, analysis=None
                ),
            ),
            (),
        ),
    ).log_request(_DEFAULT_INPUT)[1]()


def test_export_response_text():
    monitor(
        _get_mock_openai_class((_DEFAULT_RESPONSE,), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        {"export_response_texts": True},
        mona_clients_getter=get_mock_mona_clients_getter(
            (_get_mona_message(response=_DEFAULT_RESPONSE),), ()
        ),
    ).create(**_DEFAULT_INPUT)


def test_export_prompt():
    monitor(
        _get_mock_openai_class((_DEFAULT_RESPONSE,), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        {"export_prompt": True},
        mona_clients_getter=get_mock_mona_clients_getter(
            (_get_mona_message(input=_DEFAULT_INPUT),), ()
        ),
    ).create(**_DEFAULT_INPUT)


def test_multiple_answers():
    new_input = deepcopy(_DEFAULT_INPUT)
    new_input["n"] = 3
    expected_input = _remove_text_content_from_input(new_input)

    new_response = deepcopy(_DEFAULT_RESPONSE)
    new_response["choices"] = [
        {
            "finish_reason": "length",
            "index": 0,
            "logprobs": None,
            "message": {
                "role": "assistant",
                "content": "\n\nMy name is",
            },
        },
        {
            "finish_reason": "length",
            "index": 1,
            "logprobs": None,
            "message": {
                "role": "assistant",
                "content": "\n\nMy thing is",
            },
        },
        {
            "finish_reason": "length",
            "index": 2,
            "logprobs": None,
            "message": {
                "role": "assistant",
                "content": "\n\nbladf",
            },
        },
    ]

    new_expected_response = _get_response_without_texts(new_response)

    new_analysis = {
        "privacy": {
            "total_prompt_phone_number_count": 0,
            "answer_unknown_phone_number_count": (0, 0, 0),
            "total_prompt_email_count": 0,
            "answer_unknown_email_count": (0, 0, 0),
            "last_user_message_phone_number_count": 0,
            "last_user_message_emails_count": 0,
        },
        "textual": {
            "total_prompt_length": 35,
            "answer_length": (12, 13, 7),
            "total_prompt_word_count": 7,
            "answer_word_count": (3, 3, 1),
            "total_prompt_preposition_count": 2,
            "total_prompt_preposition_ratio": 0.2857142857142857,
            "answer_preposition_count": (0, 0, 0),
            "answer_preposition_ratio": (0, 0, 0),
            "answer_words_not_in_prompt_count": (3, 3, 1),
            "answer_words_not_in_prompt_ratio": (1.0, 1.0, 1.0),
            "last_user_message_length": 35,
            "last_user_message_word_count": 7,
            "last_user_message_preposition_count": 2,
            "last_user_message_preposition_ratio": 0.2857142857142857,
        },
        "profanity": {
            "prompt_profanity_prob": (0.05,),
            "prompt_has_profanity": (False,),
            "answer_profanity_prob": (0.05, 0.01, 0.05),
            "answer_has_profanity": (False, False, False),
            "last_user_message_profanity_prob": 0.05,
            "last_user_message_has_profanity": False,
        },
    }

    monitor(
        _get_mock_openai_class((new_response,), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (
                _get_mona_message(
                    response=new_expected_response,
                    input=expected_input,
                    analysis=new_analysis,
                ),
            ),
            (),
        ),
    ).create(**new_input)


def test_stream():
    def response_generator():
        words = _DEFAULT_RESPONSE_TEXT.split(" ")
        last_index = len(words) - 1
        for i, word in enumerate(words):
            choice = {
                "delta": {"content": (word + " ") if i < last_index else word},
                "index": 0,
                "logprobs": None,
                "finish_reason": None if i < last_index else "length",
            }
            yield _DEFAULT_RESPONSE_COMMON_VARIABLES | {"choices": [choice]}

    input = deepcopy(_DEFAULT_INPUT)
    input["stream"] = True

    expected_input = _remove_text_content_from_input(input)

    for _ in monitor(
        _get_mock_openai_class((response_generator(),), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (_get_mona_message(is_stream=True, input=expected_input),), ()
        ),
    ).create(**input):
        pass


def test_stream_multiple_answers():
    def response_generator():
        words = _DEFAULT_RESPONSE_TEXT.split(" ")
        for i, word in enumerate(words):
            yield _DEFAULT_RESPONSE_COMMON_VARIABLES | {
                "choices": [
                    {
                        "delta": {
                            "content": (word + " ")
                            if i < len(words) - 1
                            else word
                        },
                        "index": 0,
                        "logprobs": None,
                        "finish_reason": None
                        if i < len(words) - 1
                        else "length",
                    }
                ]
            }
            yield _DEFAULT_RESPONSE_COMMON_VARIABLES | {
                "choices": [
                    {
                        "delta": {
                            "content": (word + " ")
                            if i < len(words) - 1
                            else word
                        },
                        "index": 1,
                        "logprobs": None,
                        "finish_reason": None
                        if i < len(words) - 1
                        else "length",
                    }
                ]
            }

    input = deepcopy(_DEFAULT_INPUT)
    input["stream"] = True
    input["n"] = 2

    expected_input = _remove_text_content_from_input(input)

    expected_response = deepcopy(_DEFAULT_EXPORTED_RESPONSE)
    expected_response["choices"] += deepcopy(expected_response["choices"])
    expected_response["choices"][1]["index"] = 1
    expected_response["usage"] = {
        "completion_tokens": 8,
        "prompt_tokens": 8,
        "total_tokens": 16,
    }

    new_analysis = {
        "privacy": {
            "total_prompt_phone_number_count": 0,
            "answer_unknown_phone_number_count": (0, 0),
            "total_prompt_email_count": 0,
            "answer_unknown_email_count": (0, 0),
            "last_user_message_phone_number_count": 0,
            "last_user_message_emails_count": 0,
        },
        "textual": {
            "total_prompt_length": 35,
            "answer_length": (12, 12),
            "total_prompt_word_count": 7,
            "answer_word_count": (3, 3),
            "total_prompt_preposition_count": 2,
            "total_prompt_preposition_ratio": 0.2857142857142857,
            "answer_preposition_count": (0, 0),
            "answer_preposition_ratio": (0, 0),
            "answer_words_not_in_prompt_count": (3, 3),
            "answer_words_not_in_prompt_ratio": (1.0, 1.0),
            "last_user_message_length": 35,
            "last_user_message_word_count": 7,
            "last_user_message_preposition_count": 2,
            "last_user_message_preposition_ratio": 0.2857142857142857,
        },
        "profanity": {
            "prompt_profanity_prob": (0.05,),
            "prompt_has_profanity": (False,),
            "answer_profanity_prob": (0.05, 0.05),
            "answer_has_profanity": (False, False),
            "last_user_message_profanity_prob": 0.05,
            "last_user_message_has_profanity": False,
        },
    }

    for _ in monitor(
        _get_mock_openai_class((response_generator(),), ()),
        (),
        _DEFAULT_CONTEXT_CLASS,
        mona_clients_getter=get_mock_mona_clients_getter(
            (
                _get_mona_message(
                    is_stream=True,
                    input=expected_input,
                    response=expected_response,
                    analysis=new_analysis,
                ),
            ),
            (),
        ),
    ).create(**input):
        pass



from ..exceptions import (
    WrongOpenAIClassException,
    InvalidSamplingRatioException,
)


def validate_openai_class(openai_class, required_name):
    """
    Validates that the given OpenAI API class' name is the given
    required name.
    """
    class_name = openai_class.__name__
    if class_name != required_name:
        raise WrongOpenAIClassException(
            f"Name is {class_name} and must be {required_name}"
        )


def validate_and_get_sampling_ratio(specs):
    """
    Validates that the sampling ratio in a given specs dict is a valid
    number (between 0 and 1). Returns a default value of 1 if no
    sampling ratio is mentioned.
    """
    sampling_ratio = specs.get("sampling_ratio", 1)
    if sampling_ratio < 0 or sampling_ratio > 1:
        raise InvalidSamplingRatioException(
            f"sampling ratio is {sampling_ratio} but must be a number "
            f"between 0 and 1 (inclusive)"
        )
    return sampling_ratio



"""
The Mona wrapping code for OpenAI's Completion API.
"""
from copy import deepcopy
from functools import wraps

from ..analysis.privacy import get_privacy_analyzers
from ..analysis.profanity import get_has_profanity, get_profanity_prob
from ..analysis.textual import get_textual_analyzers
from ..util.oop_util import create_combined_object
from .endpoint_wrapping import OpenAIEndpointWrappingLogic

COMPLETION_CLASS_NAME = "Completion"


def _get_prompts(request):
    prompts = request.get("prompt", ())
    return (prompts,) if isinstance(prompts, str) else prompts


def _get_choices_texts(response):
    return tuple((choice["text"] for choice in response["choices"]))


def _get_texts(func):
    def wrapper(self, input, response):
        return func(self, _get_prompts(input), _get_choices_texts(response))

    return wrapper


def _get_analyzers(analyzers_getter):
    def decorator(func):
        @wraps(func)
        def wrapper(self, prompts, answers):
            return func(
                self, analyzers_getter(prompts), analyzers_getter(answers)
            )

        return wrapper

    return decorator


class CompletionWrapping(OpenAIEndpointWrappingLogic):
    def _get_endpoint_name(self):
        return COMPLETION_CLASS_NAME

    def get_clean_message(self, message):
        """
        Returns a copy of the given message with relevant data removed, for
        example the actual texts, to avoid sending such information, that
        is sometimes sensitive, to Mona.
        """
        new_message = deepcopy(message)
        if not self._specs.get("export_prompt", False):
            new_message["input"].pop("prompt", None)

        if "response" in message and not self._specs.get(
            "export_response_texts", False
        ):
            for choice in new_message["response"]["choices"]:
                choice.pop("text", None)

        return new_message

    @_get_texts
    @_get_analyzers(get_privacy_analyzers)
    def _get_full_privacy_analysis(
        self, prompts_privacy_analyzers, answers_privacy_analyzers
    ):
        combined_prompts = create_combined_object(prompts_privacy_analyzers)
        combined_answers = create_combined_object(answers_privacy_analyzers)
        return {
            "prompt_phone_number_count": (
                combined_prompts.get_phone_numbers_count()
            ),
            "answer_unknown_phone_number_count": (
                combined_answers.get_previously_unseen_phone_numbers_count(
                    prompts_privacy_analyzers
                )
            ),
            "prompt_email_count": combined_prompts.get_emails_count(),
            "answer_unknown_email_count": (
                combined_answers.get_previously_unseen_emails_count(
                    prompts_privacy_analyzers
                )
            ),
        }

    @_get_texts
    @_get_analyzers(get_textual_analyzers)
    def _get_full_textual_analysis(
        self, prompts_textual_analyzers, answers_textual_analyzers
    ):
        combined_prompts = create_combined_object(prompts_textual_analyzers)
        combined_answers = create_combined_object(answers_textual_analyzers)
        return {
            "prompt_length": combined_prompts.get_length(),
            "answer_length": combined_answers.get_length(),
            "prompt_word_count": combined_prompts.get_word_count(),
            "answer_word_count": combined_answers.get_word_count(),
            "prompt_preposition_count": (
                combined_prompts.get_preposition_count()
            ),
            "prompt_preposition_ratio": (
                combined_prompts.get_preposition_ratio()
            ),
            "answer_preposition_count": (
                combined_answers.get_preposition_count()
            ),
            "answer_preposition_ratio": (
                combined_answers.get_preposition_ratio()
            ),
            "answer_words_not_in_prompt_count": (
                combined_answers.get_words_not_in_others_count(
                    prompts_textual_analyzers
                )
            ),
            "answer_words_not_in_prompt_ratio": tuple(
                analyzer.get_words_not_in_others_count(
                    prompts_textual_analyzers
                )
                / analyzer.get_word_count()
                if analyzer.get_word_count() > 0
                else 0.0
                for analyzer in answers_textual_analyzers
            ),
        }

    @_get_texts
    def _get_full_profainty_analysis(self, prompts, answers):
        return {
            "prompt_profanity_prob": get_profanity_prob(prompts),
            "prompt_has_profanity": get_has_profanity(prompts),
            "answer_profanity_prob": get_profanity_prob(answers),
            "answer_has_profanity": get_has_profanity(answers),
        }

    def get_stream_delta_text_from_choice(self, choice):
        return choice["text"]

    def get_final_choice(self, text):
        return {"text": text}

    def get_all_prompt_texts(self, request):
        return _get_prompts(request)

    def get_all_response_texts(self, response):
        return _get_choices_texts(response)



from typing import Iterable

"""
Functionality for extracting privacy information from GAI responses
when comparing to given input prompts.

TODO(itai): Add many more functions to extract information such as:
    Bank accounts, full names, SSN, passport numbers, etc...
"""
import re

from phonenumbers import PhoneNumberMatcher

# TODO(itai): Add module-level tests for this module, specifically for email
#   extraction, since this is our own logic and not using an external library.

EMAIL_RE_PATTERN = (
    r"([-!#-'*+/-9=?A-Z^-~]+(\.[-!#-'*+/-9=?A-Z^-~]+)*|\"([]!#-[^-~"
    r" \t]|(\\[\t -~]))+\")@([-!#-'*+/-9=?A-Z^-~]+(\.[-!#-'*+/-9=?A"
    r"-Z^-~]+)*|\[[\t -Z^-~]*])"
)


def _extract_phone_numbers(text):
    """
    Extract phone numbers from a prompt string and return as a set.
    """
    phone_numbers = set()
    # We use "US" just as a default region in case there are no country codes
    # since we don't care about the formatting of the found number, but just
    # whether it is a phone number or not, this has no consequences.
    for match in PhoneNumberMatcher(text, "US"):
        number_string = "+{}{}".format(
            match.number.country_code, match.number.national_number
        )
        phone_numbers.add(number_string)
    return phone_numbers


def _extract_all_emails(text):
    """
    returns all email addresses found in the given prompt.
    """
    return set(re.findall(EMAIL_RE_PATTERN, text))


class PrivacyAnalyzer:
    """
    An analyzer class that takes a text and provides functionality to extract
    privacy-related metrics from that text.
    """

    def __init__(self, text) -> None:
        self._text = text
        self._phone_numbers = _extract_phone_numbers(text)
        self._emails = _extract_all_emails(text)

    def get_phone_numbers_count(self):
        """
        Returns the number of phone numbers in the initially given text.
        """
        return len(self._phone_numbers)

    def get_emails_count(self):
        """
        Returns the number of email addresses in the initially given text.
        """
        return len(self._emails)

    @classmethod
    def _get_phone_numbers_from_instance(cls, instance):
        return instance._phone_numbers

    @classmethod
    def _get_emails_from_instance(cls, instance):
        return instance._emails

    def _get_previously_unseen_x_count(
        self, others: Iterable["PrivacyAnalyzer"], extraction_function
    ):
        return len(
            extraction_function(self)
            - set().union(
                *tuple(extraction_function(other) for other in others)
            )
        )

    def get_previously_unseen_phone_numbers_count(
        self, others: Iterable["PrivacyAnalyzer"]
    ):
        """
        Returns the number of phone numbers in the initially given text, that
        don't also appear in any of the given other analyzers.
        """
        return self._get_previously_unseen_x_count(
            others, self._get_phone_numbers_from_instance
        )

    def get_previously_unseen_emails_count(
        self, others: Iterable["PrivacyAnalyzer"]
    ):
        """
        Returns the number of email addresses in the initially given text,
        that don't also appear in any of the given other analyzers.
        """
        return self._get_previously_unseen_x_count(
            others, self._get_emails_from_instance
        )


def get_privacy_analyzers(texts):
    """
    Returns a tuple of PrivacyAnalyzer objects, one for each text in the given
    iterable.
    """
    return tuple(PrivacyAnalyzer(text) for text in texts)



from .logger import Logger
import atexit
import json


class FileLogger(Logger):
    """
    A simple logging class that saves monitored data in a file.
    """

    def __init__(self, file_name):
        self.file = open(file_name, "w")

        atexit.register(self.close_file)

    def close_file(self):
        if not self.file.closed:
            self.file.close()

    def log(self, message: dict, context_id=None, export_timestamp=None):
        self.file.writelines(
            [
                json.dumps(
                    {
                        "message": message,
                        "context_id": context_id,
                        "export_timestamp": export_timestamp,
                    }
                )
            ]
        )

    async def alog(
        self, message: dict, context_id=None, export_timestamp=None
    ):
        # TODO: Imlement actual asyncio usage.
        return self.log(message, context_id, export_timestamp)
